import argparse
import asyncio

from .parser import BatteryData, InertialMeasurementData, UnknownData, WT9011Parser
from .sensor import WT9011, DeviceNotFoundError


def parse_args() -> argparse.Namespace:
    arg_parser = argparse.ArgumentParser()

    device_group = arg_parser.add_mutually_exclusive_group()

    device_group.add_argument(
        "-n",
        "--name",
        metavar="<name>",
        default="WT901BLE67",
        help="The name of the bluetooth device to connect to",
    )

    device_group.add_argument(
        "-a",
        "--address",
        metavar="<address>",
        help="The mac address of the bluetooth device to connect to",
    )

    def valid_return_rate(value):
        allowed_values = {0.2, 0.5, 1, 2, 5, 10, 20, 50, 100, 200, -1}
        try:
            value = float(value)
        except ValueError:
            pass
        if value not in allowed_values:
            raise argparse.ArgumentTypeError(
                f"Invalid return rate. Allowed values are: {', '.join(map(str, allowed_values))}."
            )
        return value

    arg_parser.add_argument(
        "-r",
        "--return-rate",
        type=valid_return_rate,
        metavar="<rate>",
        help="Set the return rate in Hz. Allowed values are 0.2, 0.5, 1, 2, 5, 10, 20, 50, 100, 200 and -1 for single return",
    )

    arg_parser.add_argument(
        "-b",
        "--battery",
        action="store_true",
        help="Get battery level",
    )

    arg_parser.add_argument(
        "-l",
        "--listen",
        action="store_true",
        help="Listen for imu notifies",
    )

    arg_parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Show unknown data frames",
    )

    return arg_parser.parse_args()


def consume(args: argparse.Namespace, data: bytearray):
    data_parser = WT9011Parser()
    types_to_parse = []

    if args.listen:
        types_to_parse.append(InertialMeasurementData)

    if args.battery:
        types_to_parse.append(BatteryData)

    if args.debug:
        types_to_parse.append(UnknownData)

    parsed_data_list = data_parser.parse(types_to_parse, data)

    for parsed in parsed_data_list:
        match parsed:
            case InertialMeasurementData():
                parsed.pretty_print()
            case BatteryData():
                print(parsed)
            case UnknownData():
                print(parsed)
                print(f"len={len(parsed.raw_data)}")
        print()


async def run_queue_consumer(args: argparse.Namespace, queue: asyncio.Queue):
    print("Starting the consumer queue")

    while True:
        # Use await asyncio.wait_for(queue.get(), timeout=1.0) if you want a timeout for getting data.
        epoch, data = await queue.get()
        if data is None:
            print("Exiting the consumer loop...")
            break
        else:
            consume(args, data)


async def ble_main(args: argparse.Namespace):
    queue = asyncio.Queue()

    sensor = WT9011(args.address, args.name)
    client_task = sensor.run_ble_client(queue, args.listen, args.battery, args.return_rate)
    consumer_task = run_queue_consumer(args, queue)

    try:
        await asyncio.gather(client_task, consumer_task)
    except DeviceNotFoundError:
        pass


def main():
    args = parse_args()
    asyncio.run(ble_main(args))



import struct
from dataclasses import dataclass

from .sensor import WT9011


@dataclass
class Vector3:
    x: float
    y: float
    z: float

    def __mul__(self, other: float):
        return Vector3(self.x * other, self.y * other, self.z * other)

    def __truediv__(self, other: float):
        return Vector3(self.x / other, self.y / other, self.z / other)


@dataclass
class InertialMeasurementData:
    lin_accel: Vector3  # linear acceleration
    ang_vel: Vector3  # angular velocity
    angle: Vector3

    def pretty_print(self):
        print(
            "Linear Acceleration (m/s^2):    "
            f"   ax = {self.lin_accel.x:>8.2f}  "
            f"   ay = {self.lin_accel.y:>8.2f}  "
            f"   az = {self.lin_accel.z:>8.2f}"
        )
        print(
            "Angular Velocity (°/s):         "
            f"   wx = {self.ang_vel.x:>8.2f}  "
            f"   wy = {self.ang_vel.y:>8.2f}  "
            f"   wz = {self.ang_vel.z:>8.2f}"
        )
        print(
            "Angle (°):                      "
            f" roll = {self.angle.x:>8.2f}  "
            f"pitch = {self.angle.y:>8.2f}  "
            f"  yaw = {self.angle.z:>8.2f}"
        )


@dataclass
class BatteryData:
    voltage: float
    percentage: float


@dataclass
class UnknownData:
    raw_data: bytearray


# WT9011 Datasheet/Manual src:
# https://manuals.plus/m/65ba8c4db28ab28226fb0dc1667166259ea316332b848d51c296430336998cb2_optim.pdf
class WT9011Parser:
    def parse(self, types_to_parse: [], data: bytearray) -> []:
        parsed_data_list = []

        # Frames are 20-byte raw data packets sent by the sensor
        # The first byte is a packet header
        # and the second byte is a flag byte
        # Whole definition is in the docs (src: page 15.)
        frames = self._split_to_frames(data)
        for frame in frames:
            parsed_data = self._parse_frame(types_to_parse, frame)
            if parsed_data is not None:
                parsed_data_list.append(parsed_data)
        return parsed_data_list

    def _split_to_frames(self, data: bytearray) -> []:
        delimiter = bytearray([WT9011.PACKET_HEADER])

        # Split data array to frames
        frames = []
        # split_parts are frames without the packet packet_header
        # data.split() removes the delimiter from the array
        split_parts = data.split(delimiter)
        for part in split_parts:
            # When the delimiter is the first element of data
            # data.split() adds an b'' as a first element of split_parts
            # Skip it as it doesn't contain any data
            if part == b"":
                continue

            # Re-add the packet header to each frame
            frame = delimiter + part
            frames.append(frame)
        return frames

    def _parse_frame(self, types_to_parse: [], frame):
        if len(frame) != 20:
            return UnknownData(frame) if UnknownData in types_to_parse else None

        flag_byte = frame[1]
        match flag_byte:
            case WT9011.FLAG_INERTIAL:
                return self._parse_imu(frame) if InertialMeasurementData in types_to_parse else None
            case WT9011.FLAG_BATTERY:
                return self._parse_battery(frame) if BatteryData in types_to_parse else None
            case _:
                return UnknownData(frame) if UnknownData in types_to_parse else None

    def _parse_imu(self, frame: bytearray) -> InertialMeasurementData:
        # B- byte
        # h- short (2 bytes)
        fmt = "BBhhhhhhhhh"
        (packet_header, mark_bit, ax, ay, az, wx, wy, wz, roll, pitch, yaw) = struct.unpack(
            fmt, frame
        )

        EARTH_ACCELERATION = 9.8  # in m/s^2

        # Measurement ranges of the sensors (src: page 7.)
        ACCEL_RANGE = 16.0  # in g units
        GYRO_RANGE = 2000.0  # in degrees/s
        ANGLE_RANGE = 180.0  # in degrees

        # Used to normalize the data from sensor values to the (-1, 1) range
        # Registers are 8-bits long, two registers combined (16-bits) store a sensor value
        # 16-bit signed integers range from -32768 to 32767
        # To better understand what this variable represents see *the example below*
        SCALING_FACTOR = 32768.0

        # *the example below*
        # ax - 16 bit sensor data in range (-32768, 32767)
        # ax / SCALING_FACTOR - sensor data normalized to range (-1, 1)
        # ax / SCALING_FACTOR * ACCEL_RANGE * GRAVITY_ACCELERATION - normalized scaled to the real world units based on the sensor range
        lin_accel = Vector3(ax, ay, az) / SCALING_FACTOR * ACCEL_RANGE * EARTH_ACCELERATION

        ang_vel = Vector3(wx, wy, wz) / SCALING_FACTOR * GYRO_RANGE

        angle = Vector3(roll, pitch, yaw) / SCALING_FACTOR * ANGLE_RANGE

        return InertialMeasurementData(lin_accel, ang_vel, angle)

    def _parse_battery(self, frame: bytearray) -> BatteryData:
        MAX_VOLTAGE = 4.1
        MIN_VOLTAGE = 3.2
        voltage_raw = int.from_bytes(frame[4:6], byteorder="little")
        voltage = voltage_raw / 100.0

        def map_range(value, in_min, in_max, out_min, out_max):
            mapped = (value - in_min) * (out_max - out_min) / (in_max - in_min) + out_min
            return max(out_min, min(out_max, mapped))  # Clamp to [out_min, out_max]

        percentage = map_range(voltage, MIN_VOLTAGE, MAX_VOLTAGE, 0, 100)

        return BatteryData(voltage, percentage)



import asyncio
import time

from bleak import BleakClient, BleakScanner


class DeviceNotFoundError(Exception):
    pass


class WT9011:
    # GATT characteristics:
    # Notify characteristic UUID
    CHAR_NOTIFY = "0000ffe4-0000-1000-8000-00805f9a34fb"
    # Write characteristic UUID
    CHAR_WRITE = "0000ffe9-0000-1000-8000-00805f9a34fb"

    DESCRIPTOR = "00002902-0000-1000-8000-00805f9b34fb"

    # Request battery info
    REQ_BAT_INFO = bytearray([0xFF, 0xAA, 0x27, 0x64, 0x00])

    # Header and flags:
    PACKET_HEADER = 0x55
    # Flag in a frame containing inertial data
    FLAG_INERTIAL = 0x61
    # Flag in a frame containing battery data
    FLAG_BATTERY = 0x71

    def __init__(self, address, name):
        if address is None and name is None:
            raise Exception("Either address or name must be specified to create a sensor")
        self.address = address
        self.name = name

    async def _find_device(self):
        print("Starting the scan...")
        if self.address:
            device = await BleakScanner.find_device_by_address(self.address)
            if device is None:
                print(f"Could not find the device with address '{self.address}'")
                raise DeviceNotFoundError
        else:
            device = await BleakScanner.find_device_by_name(self.name)
            if device is None:
                print(f"Could not find the device with name '{self.name}")
                raise DeviceNotFoundError
        return device

    async def _listen_for_notifies(self, queue: asyncio.Queue, client: BleakClient):

        # TODO: make callback_handler a class method?
        # TODO: make queue a class field?
        async def callback_handler(_, data):
            await queue.put((time.time(), data))

        # TODO: run until stopped with ctrl-c instead of running 10 seconds
        print("Listening for notifies for 10 seconds")
        await client.start_notify(WT9011.CHAR_NOTIFY, callback_handler)
        await asyncio.sleep(10.0)
        await client.stop_notify(WT9011.CHAR_NOTIFY)

    async def _get_battery_level(self, queue: asyncio.Queue, client: BleakClient):

        async def callback_handler(_, data):
            await queue.put((time.time(), data))

        print("Getting battery level")
        # Using this notify hack instead of getting a proper response from write (response = await client.write_gatt_char())
        # because of this issue https://github.com/hbldh/bleak/issues/59
        await client.start_notify(WT9011.CHAR_NOTIFY, callback_handler)
        await client.write_gatt_char(WT9011.CHAR_WRITE, self.REQ_BAT_INFO, response=True)
        await asyncio.sleep(0.5)  # Sleeping just to make sure the response is not missed...
        await client.stop_notify(WT9011.CHAR_NOTIFY)

    async def _set_return_rate(self, client: BleakClient, rate: float):
        return_rate_to_hex = {
            0.2: 0x01,
            0.5: 0x02,
            1.0: 0x03,
            2.0: 0x04,
            5.0: 0x05,
            10.0: 0x06,
            20.0: 0x07,
            50.0: 0x08,
            100.0: 0x09,
            200.0: 0x0B,
            -1.0: 0x0C,
        }
        hex = return_rate_to_hex.get(rate, None)
        if hex is None:
            raise ValueError("Wrong return rate")

        command = bytearray([0xFF, 0xAA, 0x03, hex, 0x00])
        print(f"Setting the return rate to {'single return' if rate == -1 else str(rate) + ' Hz'}")
        await client.write_gatt_char(WT9011.CHAR_WRITE, command, response=False)

    async def run_ble_client(
        self, queue: asyncio.Queue, listen=True, battery=False, return_rate=None
    ):
        device = await self._find_device()
        print("Connecting to the device...")

        async with BleakClient(device) as client:
            print("Connected")

            if return_rate is not None:
                await self._set_return_rate(client, return_rate)

            if battery:
                await self._get_battery_level(queue, client)

            if listen:
                await self._listen_for_notifies(queue, client)

            # Send an "exit command to the consumer"
            await queue.put((time.time(), None))

        print("Disconnected")
